function [A, W] = fpica(X, whiteningMatrix, dewhiteningMatrix, approach, numOfIC, g, finetune, a1, a2, myy, stabilization, epsilon, maxNumIterations, maxFinetune, initState, guess, sampleSize, displayMode, displayInterval, s_verbose);
//  Fixed point ICA. Main algorithm of FASTICA.
// Calling Sequence
// [A, W] = fpica(whitesig, whiteningMatrix, dewhiteningMatrix, approach, numOfIC, g, finetune, a1, a2, mu, stabilization, epsilon, maxNumIterations, maxFinetune, initState, guess, sampleSize, displayMode, displayInterval, verbose);
// Parameters
// whitesig                              :the whitened data as row vectors
// whiteningMatrix                       :can be obtained with function whitenv
// dewhiteningMatrix                     :can be obtained with function whitenv
// approach           : [ 'symm' | 'defl' ] the approach used (deflation or symmetric)
// numOfIC        : [ 0 - Dim of whitesig ] number of independent components estimated
// g                 :  [ 'pow3' | 'tanh' |  'gaus' | 'skew' ] the nonlinearity used
// finetune          :  [same as g + 'off'] the nonlinearity used in finetuning.
// a1                                    :parameter for tuning 'tanh'
// a2                                    :parameter for tuning 'gaus'
// mu                                    :step size in stabilized algorithm
// stabilization       :  [ 'on' | 'off' ]  if mu < 1 then automatically on
// epsilon                               :stopping criterion
// maxNumIterations                      :maximum number of iterations 
// maxFinetune                           :maximum number of iteretions for finetuning
// initState         : [ 'rand' | 'guess' ] initial guess or random initial state. See below
// guess                                 :initial guess for A. Ignored if initState = 'rand'
// sampleSize                   : [ 0 - 1 ] percentage of the samples used in one iteration
// displayMode    : [ 'signals' | 'basis' | 'filters' | 'off' ] plot running estimate
// displayInterval                       :number of iterations we take between plots
// verbose              : [ 'on' | 'off' ]  report progress in text format
//
// Description
// Perform independent component analysis using Hyvarinen's fixed point
// algorithm. Outputs an estimate of the mixing matrix A and its inverse W.
// Examples
//       [E, D] = pcamat(vectors);
//       [nv, wm, dwm] = whitenv(vectors, E, D);
//       [A, W] = fpica(nv, wm, dwm);
//
//   See also
//   fastica
//   fasticag
//   whitenv
//   pcamat
//   Authors
//   H. Nahrstaedt - 2010
//   (C) 1996-2005 by Hugo Gävert, Jarmo Hurri, Jaakko Särelä, and Aapo Hyvärinen. 

// @(#)$Id: fpica.m,v 1.7 2005/06/16 12:52:55 jarmo Exp $

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Global variable for stopping the ICA calculations from the GUI
global g_FastICA_interrupt;
if isempty(g_FastICA_interrupt)
  clear  g_FastICA_interrupt;
  interruptible = 0;
else
  interruptible = 1;
end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Default values

if argn(2) < 3, error('Not enough arguments!'); end
[vectorSize, numSamples] = size(X);
if argn(2) < 20, s_verbose = 'on'; end
if argn(2) < 19, displayInterval = 1; end
if argn(2) < 18, displayMode = 'on'; end
if argn(2) < 17, sampleSize = 1; end
if argn(2) < 16, guess = 1; end
if argn(2) < 15, initState = 'rand'; end
if argn(2) < 14, maxFinetune = 100; end
if argn(2) < 13, maxNumIterations = 1000; end
if argn(2) < 12, epsilon = 0.0001; end
if argn(2) < 11, stabilization = 'on'; end
if argn(2) < 10, myy = 1; end
if argn(2) < 9, a2 = 1; end
if argn(2) < 8, a1 = 1; end
if argn(2) < 7, finetune = 'off'; end
if argn(2) < 6, g = 'pow3'; end
if argn(2) < 5, numOfIC = vectorSize; end     // vectorSize = Dim
if argn(2) < 4, approach = 'defl'; end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Checking the data

if ~isreal(X)
  error('Input has an imaginary part.');
end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Checking the value for verbose
select convstr(s_verbose,"l")
 case 'on' then
  b_verbose = 1;
 case 'off' then
  b_verbose = 0;
 else 
  error(sprintf('Illegal value [ %s ] for parameter: ''verbose''\n', s_verbose));
end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Checking the value for approach
select convstr(approach,"l")
 case 'symm' then
  approachMode = 1;
 case 'defl' then
  approachMode = 2;
 else
  error(sprintf('Illegal value [ %s ] for parameter: ''approach''\n', approach));
end
if b_verbose, printf('Used approach [ %s ].\n', approach); end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Checking the value for numOfIC

if vectorSize < numOfIC
  error('Must have numOfIC <= Dimension!');
end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Checking the sampleSize
if sampleSize > 1
  sampleSize = 1;
  if b_verbose
    printf('Warning: Setting ''sampleSize'' to 1.\n');
  end  
elseif sampleSize < 1
  if (sampleSize * numSamples) < 1000
    sampleSize = min(1000/numSamples, 1);
    if b_verbose
      printf('Warning: Setting ''sampleSize'' to %0.3f (%d samples).\n', ...
	      sampleSize, floor(sampleSize * numSamples));
    end  
  end
end
if b_verbose
  if  b_verbose & (sampleSize < 1)
    printf('Using about %0.0f%% of the samples in random order in every step.\n',sampleSize*100);
  end
end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Checking the value for nonlinearity.
select convstr(g,"l")
 case 'pow3' then
  gOrig = 10;
 case 'tanh' then
  gOrig = 20;
 case 'gaus' then
  gOrig = 30;
 case 'gauss' then
  gOrig = 30;
 case 'skew' then
  gOrig = 40;
 else
  error(sprintf('Illegal value [ %s ] for parameter: ''g''\n', g));
end
if sampleSize ~= 1
  gOrig = gOrig + 2;
end
if myy ~= 1
  gOrig = gOrig + 1;
end

if b_verbose,
  printf('Used nonlinearity [ %s ].\n', g);
end

finetuningEnabled = 1;
select convstr(finetune,"l")
 case 'pow3' then
  gFine = 10 + 1;
 case 'tanh' then
  gFine = 20 + 1;
 case 'gaus' then
  gFine = 30 + 1;
 case  'gauss' then
  gFine = 30 + 1;
 case 'skew' then
  gFine = 40 + 1;
 case 'off' then
  if myy ~= 1
    gFine = gOrig;
  else 
    gFine = gOrig + 1;
  end
  finetuningEnabled = 0;
 else
  error(sprintf('Illegal value [ %s ] for parameter: ''finetune''\n', ...
		finetune));
end

if b_verbose & finetuningEnabled
  printf('Finetuning enabled (nonlinearity: [ %s ]).\n', finetune);
end

select convstr(stabilization,"l")
 case 'on' then
  stabilizationEnabled = 1;
 case 'off' then
  if myy ~= 1
    stabilizationEnabled = 1;
  else
    stabilizationEnabled = 0;
  end
 else
  error(sprintf('Illegal value [ %s ] for parameter: ''stabilization''\n', ...
		stabilization)); 
end

if b_verbose & stabilizationEnabled
  printf('Using stabilized algorithm.\n');
end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Some other parameters
myyOrig = myy;
// When we start fine-tuning we'll set myy = myyK * myy
myyK = 0.01;
// How many times do we try for convergence until we give up.
failureLimit = 5;


usedNlinearity = gOrig;
stroke = 0;
notFine = 1;
long = 0;


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Checking the value for initial state.

select convstr(initState,"l")
 case 'rand' then
  initialStateMode = 0;
 case 'guess' then
  if size(guess,1) ~= size(whiteningMatrix,2)
    initialStateMode = 0;
    if b_verbose
      printf('Warning: size of initial guess is incorrect. Using random initial guess.\n');
    end
  else
    initialStateMode = 1;
    if size(guess,2) < numOfIC
      if b_verbose
	printf('Warning: initial guess only for first %d components. Using random initial guess for others.\n', size(guess,2)); 
      end
      guess(:, size(guess, 2) + 1:numOfIC) = ...
					     rand(vectorSize,numOfIC-size(guess,2))-.5;
    elseif size(guess,2)>numOfIC
      guess=guess(:,1:numOfIC);
      printf('Warning: Initial guess too large. The excess column are dropped.\n');
    end
    if b_verbose, printf('Using initial guess.\n'); end
  end
 else
  error(sprintf('Illegal value [ %s ] for parameter: ''initState''\n', initState));
end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Checking the value for display mode.

select convstr(displayMode,"l")
 case 'off' then
  usedDisplay = 0;
 case  'none' then
  usedDisplay = 0;
 case 'on' then
  usedDisplay = 1;
  if (b_verbose & (numSamples > 10000))
    printf('Warning: Data vectors are very long. Plotting may take long time.\n');
  end
  if (b_verbose & (numOfIC > 25))
    printf('Warning: There are too many signals to plot. Plot may not look good.\n');
  end
 case 'signals' then
  usedDisplay = 1;
  if (b_verbose & (numSamples > 10000))
    printf('Warning: Data vectors are very long. Plotting may take long time.\n');
  end
  if (b_verbose & (numOfIC > 25))
    printf('Warning: There are too many signals to plot. Plot may not look good.\n');
  end
 case 'basis'
  usedDisplay = 2;
  if (b_verbose & (numOfIC > 25))
    printf('Warning: There are too many signals to plot. Plot may not look good.\n');
  end
 case 'filters'
  usedDisplay = 3;
  if (b_verbose & (vectorSize > 25))
    printf('Warning: There are too many signals to plot. Plot may not look good.\n');
  end
 else
  error(sprintf('Illegal value [ %s ] for parameter: ''displayMode''\n', displayMode));
end

// The displayInterval can't be less than 1...
if displayInterval < 1
  displayInterval = 1;
end


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if b_verbose, printf('Starting ICA calculation...\n'); end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SYMMETRIC APPROACH
if approachMode == 1,

  // set some parameters more...
  usedNlinearity = gOrig;
  stroke = 0;
  notFine = 1;
  long = 0;
  
  A = zeros(vectorSize, numOfIC);  // Dewhitened basis vectors.
  if initialStateMode == 0
    // Take random orthonormal initial vectors.
    B = orth (rand (vectorSize, numOfIC,"normal"));
  elseif initialStateMode == 1
    // Use the given initial vector as the initial state
    B = whiteningMatrix * guess;
  end
  
  BOld = zeros(B);
  BOld2 = zeros(B);
  
  // This is the actual fixed-point iteration loop.
  for rnd = 1:maxNumIterations + 1,
    if rnd == maxNumIterations + 1,
      printf('No convergence after %d steps\n', maxNumIterations);
      printf('Note that the plots are probably wrong.\n');
      if ~isempty(B)
	// Symmetric orthogonalization.
	B = B * real(inv(B' * B)^(1/2));

	W = B' * whiteningMatrix;
	A = dewhiteningMatrix * B;
      else
	W = [];
	A = [];
      end
      return;
    end
    
    if (interruptible & g_FastICA_interrupt)
      if b_verbose 
        printf('\n\nCalculation interrupted by the user\n');
      end
      if ~isempty(B)
	W = B' * whiteningMatrix;
	A = dewhiteningMatrix * B;
      else
	W = [];
	A = [];
      end
      return;
    end
    
    
    // Symmetric orthogonalization.
    B = B * real(inv(B' * B)^(1/2));
    
    // Test for termination condition. Note that we consider opposite
    // directions here as well.
    minAbsCos = mtlb_min(abs(diag(B' * BOld)));
    minAbsCos2 = mtlb_min(abs(diag(B' * BOld2)));
    
    if (1 - minAbsCos < epsilon)
      if finetuningEnabled & notFine
        if b_verbose, printf('Initial convergence, fine-tuning: \n'); end;
        notFine = 0;
        usedNlinearity = gFine;
        myy = myyK * myyOrig;
        BOld = zeros(B);
        BOld2 = zeros(B);
	
      else
        if b_verbose, printf('Convergence after %d steps\n', rnd); end
	
        // Calculate the de-whitened vectors.
        A = dewhiteningMatrix * B;
        break;
      end
    elseif stabilizationEnabled
      if (~stroke) & (1 - minAbsCos2 < epsilon)
	if b_verbose, printf('Stroke!\n'); end;
	stroke = myy;
	myy = .5*myy;
	if pmodulo(usedNlinearity,2) == 0
	  usedNlinearity = usedNlinearity + 1;
	end
      elseif stroke
	myy = stroke;
	stroke = 0;
	if (myy == 1) & (pmodulo(usedNlinearity,2) ~= 0)
	  usedNlinearity = usedNlinearity - 1;
	end
      elseif (~long) & (rnd>maxNumIterations/2)
	if b_verbose, printf('Taking long (reducing step size)\n'); end;
	long = 1;
	myy = .5*myy;
	if pmodulo(usedNlinearity,2) == 0
	  usedNlinearity = usedNlinearity + 1;
	end
      end
    end
    
    BOld2 = BOld;
    BOld = B;
    
    //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    // Show the progress...
    if b_verbose
      if rnd == 1
        printf('Step no. %d\n', rnd);
      else
        printf('Step no. %d, change in value of estimate: %.3g \n', rnd, 1 - minAbsCos);
      end
    end
    
    //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    // Also plot the current state...
    select usedDisplay
     case 1 then
      if rem(rnd, displayInterval) == 0,
	// There was and may still be other displaymodes...
	// 1D signals
	icaplot('dispsig',(X'*B)');
	//drawnow;
      end
     case 2 then
      if rem(rnd, displayInterval) == 0,
	// ... and now there are :-)
	// 1D basis
	A = dewhiteningMatrix * B;
	icaplot('dispsig',A');
	//drawnow;
      end
     case 3 then
      if rem(rnd, displayInterval) == 0,
	// ... and now there are :-)
	// 1D filters
	W = B' * whiteningMatrix;
	icaplot('dispsig',W);
	//drawnow;
      end
     else
    end
    
    select usedNlinearity
      // pow3
     case 10 then
      B = (X * (( X' * B) .^ 3)) / numSamples - 3 * B;
     case 11 then
      // optimoitu - epsilonin kokoisia eroja
      // t on optimoitu koodi, katso vanha koodi esim.
      // aikaisemmista versioista kuten 2.0 beta3
      Y = X' * B;
      Gpow3 = Y .^ 3;
      Beta = mtlb_sum(Y .* Gpow3);
      D = diag(1 ./ (Beta - 3 * numSamples));
      B = B + myy * B * (Y' * Gpow3 - diag(Beta)) * D;
     case 12 then
      Xsub=X(:, getSamples(numSamples, sampleSize));
      B = (Xsub * (( Xsub' * B) .^ 3)) / size(Xsub,2) - 3 * B;
     case 13 then
      // Optimoitu
      Ysub=X(:, getSamples(numSamples, sampleSize))' * B;
      Gpow3 = Ysub .^ 3;
      Beta = mtlb_sum(Ysub .* Gpow3);
      D = diag(1 ./ (Beta - 3 * size(Ysub', 2)));
      B = B + myy * B * (Ysub' * Gpow3 - diag(Beta)) * D;
      
      // tanh
     case 20 then
      hypTan = my_tanh(a1 * X' * B);
      B = X * hypTan / numSamples - ones(size(B,1),1) * mtlb_sum(1 - hypTan .^ 2) .* B / numSamples *  a1;
     case 21 then
      // optimoitu - epsilonin kokoisia 
      Y = X' * B;
      hypTan = my_tanh(a1 * Y);
      Beta = mtlb_sum(Y .* hypTan);
      D = diag(1 ./ (Beta - a1 * mtlb_sum(1 - hypTan .^ 2)));
      B = B + myy * B * (Y' * hypTan - diag(Beta)) * D;
     case 22 then
      Xsub=X(:, getSamples(numSamples, sampleSize));
      hypTan = my_tanh(a1 * Xsub' * B);
      B = Xsub * hypTan / size(Xsub, 2) -  ones(size(B,1),1) * mtlb_sum(1 - hypTan .^ 2) .* B / size(Xsub, 2) * a1;
     case 23 then
      // Optimoitu
      Y = X(:, getSamples(numSamples, sampleSize))' * B;
      hypTan = my_tanh(a1 * Y);
      Beta = mtlb_sum(Y .* hypTan);
      D = diag(1 ./ (Beta - a1 * mtlb_sum(1 - hypTan .^ 2)));
      B = B + myy * B * (Y' * hypTan - diag(Beta)) * D;
      
      // gauss
     case 30 then
      U = X' * B;
      Usquared=U .^ 2;
      ex = exp(-a2 * Usquared / 2);
      gauss =  U .* ex;
      dGauss = (1 - a2 * Usquared) .*ex;
      B = X * gauss / numSamples -  ones(size(B,1),1) * mtlb_sum(dGauss) .* B / numSamples ;
     case 31 then
      // optimoitu
      Y = X' * B;
      ex = exp(-a2 * (Y .^ 2) / 2);
      gauss = Y .* ex;
      Beta = mtlb_sum(Y .* gauss);
      D = diag(1 ./ (Beta - mtlb_sum((1 - a2 * (Y .^ 2)) .* ex)));
      B = B + myy * B * (Y' * gauss - diag(Beta)) * D;
     case 32 then
      Xsub=X(:, getSamples(numSamples, sampleSize));
      U = Xsub' * B;
      Usquared=U .^ 2;
      ex = exp(-a2 * Usquared / 2);
      gauss =  U .* ex;
      dGauss = (1 - a2 * Usquared) .*ex;
      B = Xsub * gauss / size(Xsub,2) - ones(size(B,1),1) * mtlb_sum(dGauss) .* B / size(Xsub,2) ;
     case 33 then
      // Optimoitu
      Y = X(:, getSamples(numSamples, sampleSize))' * B;
      ex = exp(-a2 * (Y .^ 2) / 2);
      gauss = Y .* ex;
      Beta = mtlb_sum(Y .* gauss);
      D = diag(1 ./ (Beta - mtlb_sum((1 - a2 * (Y .^ 2)) .* ex)));
      B = B + myy * B * (Y' * gauss - diag(Beta)) * D;
      
      // skew
     case 40 then
      B = (X * ((X' * B) .^ 2)) / numSamples;
     case 41 then
      // Optimoitu
      Y = X' * B;
      Gskew = Y .^ 2;
      Beta = mtlb_sum(Y .* Gskew);
      D = diag(1 ./ (Beta));
      B = B + myy * B * (Y' * Gskew - diag(Beta)) * D;
     case 42 then
      Xsub=X(:, getSamples(numSamples, sampleSize));
      B = (Xsub * ((Xsub' * B) .^ 2)) / size(Xsub,2);
     case 43 then
      // Uusi optimoitu
      Y = X(:, getSamples(numSamples, sampleSize))' * B;
      Gskew = Y .^ 2;
      Beta = mtlb_sum(Y .* Gskew);
      D = diag(1 ./ (Beta));
      B = B + myy * B * (Y' * Gskew - diag(Beta)) * D;

     else
      error('Code for desired nonlinearity not found!');
    end
  end

  
  // Calculate ICA filters.
  W = B' * whiteningMatrix;
  
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // Also plot the last one...
  select usedDisplay
   case 1 
    // There was and may still be other displaymodes...
    // 1D signals
    icaplot('dispsig',(X'*B)');
    //drawnow;
   case 2
    // ... and now there are :-)
    // 1D basis
    icaplot('dispsig',A');
    //drawnow;
   case 3
    // ... and now there are :-)
    // 1D filters
    icaplot('dispsig',W);
    //drawnow;
   else
  end
end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// DEFLATION APPROACH
if approachMode == 2
  
  B = mtlb_zeros(vectorSize);
  
  // The search for a basis vector is repeated numOfIC times.
  rnd = 1;
  
  numFailures = 0;
  
  while rnd <= numOfIC,
    myy = myyOrig;
    usedNlinearity = gOrig;
    stroke = 0;
    notFine = 1;
    long = 0;
    endFinetuning = 0;
    
    //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    // Show the progress...
    if b_verbose, printf('IC %d ', rnd); end
    
    // Take a random initial vector of lenght 1 and orthogonalize it
    // with respect to the other vectors.
    if initialStateMode == 0
      w = rand (vectorSize, 1,"normal");
    elseif initialStateMode == 1
      w=whiteningMatrix*guess(:,rnd);
    end
    w = w - B * B' * w;
    w = w / norm(w);
    
    wOld = zeros(w);
    wOld2 = zeros(w);
    
    // This is the actual fixed-point iteration loop.
    //    for i = 1 : maxNumIterations + 1
    i = 1;
    gabba = 1;
    while i <= maxNumIterations + gabba
      if (usedDisplay > 0)
	//drawnow;
      end
      if (interruptible & g_FastICA_interrupt)
        if b_verbose 
          printf('\n\nCalculation interrupted by the user\n');
        end
        return;
      end
      
      // Project the vector into the space orthogonal to the space
      // spanned by the earlier found basis vectors. Note that we can do
      // the projection with matrix B, since the zero entries do not
      // contribute to the projection.
      w = w - B * B' * w;
      w = w / norm(w);
      
      if notFine
	if i == maxNumIterations + 1
	  if b_verbose
	    printf('\nComponent number %d did not converge in %d iterations.\n', rnd, maxNumIterations);
	  end
	  rnd = rnd - 1;
	  numFailures = numFailures + 1;
	  if numFailures > failureLimit
	    if b_verbose
	      printf('Too many failures to converge (%d). Giving up.\n', numFailures);
	    end
	    if rnd == 0
	      A=[];
	      W=[];
	    end
	    return;
	  end
	  // numFailures > failurelimit
	  break;
	end
	// i == maxNumIterations + 1
      else
	// if notFine
	if i >= endFinetuning
	  wOld = w; // So the algorithm will stop on the next test...
	end
      end
      // if notFine
      
      //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      // Show the progress...
      if b_verbose, printf('.'); end;
      
      
      // Test for termination condition. Note that the algorithm has
      // converged if the direction of w and wOld is the same, this
      // is why we test the two cases.
      if norm(w - wOld) < epsilon | norm(w + wOld) < epsilon
        if finetuningEnabled & notFine
          if b_verbose, printf('Initial convergence, fine-tuning: '); end;
          notFine = 0;
	  gabba = maxFinetune;
          wOld = mtlb_zeros(size(w));
          wOld2 = mtlb_zeros(size(w));
          usedNlinearity = gFine;
          myy = myyK * myyOrig;
	  
	  endFinetuning = maxFinetune + i;
	  
        else
          numFailures = 0;
          // Save the vector
          B(:, rnd) = w;
	  
          // Calculate the de-whitened vector.
          A(:,rnd) = dewhiteningMatrix * w;
          // Calculate ICA filter.
          W(rnd,:) = w' * whiteningMatrix;
	  
          //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
          // Show the progress...
          if b_verbose, printf('computed ( %d steps ) \n', i); end
	  
          //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
          // Also plot the current state...
          select usedDisplay
	   case 1
	    if rem(rnd, displayInterval) == 0,
	      // There was and may still be other displaymodes...   
	      // 1D signals
	      temp = X'*B;
	      icaplot('dispsig',temp(:,1:numOfIC)');
	     // drawnow;
	    end
	   case 2
	    if rem(rnd, displayInterval) == 0,
	      // ... and now there are :-) 
	      // 1D basis
	      icaplot('dispsig',A');
	      //drawnow;
	    end
	   case 3
	    if rem(rnd, displayInterval) == 0,
	      // ... and now there are :-) 
	      // 1D filters
	      icaplot('dispsig',W);
	      //drawnow;
	    end
          end
	  // switch usedDisplay
	  break; // IC ready - next...
        end
	//if finetuningEnabled & notFine
      elseif stabilizationEnabled
	if (~stroke) & (norm(w - wOld2) < epsilon | norm(w + wOld2) < ...
			epsilon)
	  stroke = myy;
	  if b_verbose, printf('Stroke!'); end;
	  myy = .5*myy;
	  if pmodulo(usedNlinearity,2) == 0
	    usedNlinearity = usedNlinearity + 1;
	  end
	elseif stroke
	  myy = stroke;
	  stroke = 0;
	  if (myy == 1) & (pmodulo(usedNlinearity,2) ~= 0)
	    usedNlinearity = usedNlinearity - 1;
	  end
	elseif (notFine) & (~long) & (i > maxNumIterations / 2)
	  if b_verbose, printf('Taking long (reducing step size) '); end;
	  long = 1;
	  myy = .5*myy;
	  if pmodulo(usedNlinearity,2) == 0
	    usedNlinearity = usedNlinearity + 1;
	  end
	end
      end
      
      wOld2 = wOld;
      wOld = w;
      
      select usedNlinearity
	// pow3
       case 10 then
	w = (X * ((X' * w) .^ 3)) / numSamples - 3 * w;
       case 11 then
	EXGpow3 = (X * ((X' * w) .^ 3)) / numSamples;
	Beta = w' * EXGpow3;
	w = w - myy * (EXGpow3 - Beta * w) / (3 - Beta);
       case 12 then
	Xsub=X(:,getSamples(numSamples, sampleSize));
	w = (Xsub * ((Xsub' * w) .^ 3)) / size(Xsub, 2) - 3 * w;
       case 13 then
	Xsub=X(:,getSamples(numSamples, sampleSize));
	EXGpow3 = (Xsub * ((Xsub' * w) .^ 3)) / size(Xsub, 2);
	Beta = w' * EXGpow3;
	w = w - myy * (EXGpow3 - Beta * w) / (3 - Beta);
	// tanh
       case 20 then
	hypTan = my_tanh(a1 * X' * w);
	w = (X * hypTan - a1 * mtlb_sum(1 - hypTan .^ 2)' * w) / numSamples;
       case 21 then
	hypTan = my_tanh(a1 * X' * w);
	Beta = w' * X * hypTan;
	w = w - myy * ((X * hypTan - Beta * w) / (a1 * mtlb_sum((1-hypTan .^2)') - Beta));
       case 22 then
	Xsub=X(:,getSamples(numSamples, sampleSize));
	hypTan = my_tanh(a1 * Xsub' * w);
	w = (Xsub * hypTan - a1 * mtlb_sum(1 - hypTan .^ 2)' * w) / size(Xsub, 2);
       case 23 then
	Xsub=X(:,getSamples(numSamples, sampleSize));
	hypTan = my_tanh(a1 * Xsub' * w);
	Beta = w' * Xsub * hypTan;
	w = w - myy * ((Xsub * hypTan - Beta * w) /  (a1 * mtlb_sum((1-hypTan .^2)') - Beta));
	// gauss
       case 30 then
	// This has been split for performance reasons.
	u = X' * w;
	u2=u.^2;
	ex=exp(-a2 * u2/2);
	gauss =  u.*ex;
	dGauss = (1 - a2 * u2) .*ex;
	w = (X * gauss - mtlb_sum(dGauss)' * w) / numSamples;
       case 31 then
	u = X' * w;
	u2=u.^2;
	ex=exp(-a2 * u2/2);
	gauss =  u.*ex;
	dGauss = (1 - a2 * u2) .*ex;
	Beta = w' * X * gauss;
	w = w - myy * ((X * gauss - Beta * w) /   (mtlb_sum(dGauss)' - Beta));
       case 32 then
	Xsub=X(:,getSamples(numSamples, sampleSize));
	u = Xsub' * w;
	u2=u.^2;
	ex=exp(-a2 * u2/2);
	gauss =  u.*ex;
	dGauss = (1 - a2 * u2) .*ex;
	w = (Xsub * gauss - mtlb_sum(dGauss)' * w) / size(Xsub, 2);
       case 33 then
	Xsub=X(:,getSamples(numSamples, sampleSize));
	u = Xsub' * w;
	u2=u.^2;
	ex=exp(-a2 * u2/2);
	gauss =  u.*ex;
	dGauss = (1 - a2 * u2) .*ex;
	Beta = w' * Xsub * gauss;
	w = w - myy * ((Xsub * gauss - Beta * w) /  (mtlb_sum(dGauss)' - Beta));
	// skew
       case 40 then 
	w = (X * ((X' * w) .^ 2)) / numSamples;
       case 41 then
	EXGskew = (X * ((X' * w) .^ 2)) / numSamples;
	Beta = w' * EXGskew;
	w = w - myy * (EXGskew - Beta*w)/(-Beta);
       case 42 then
	Xsub=X(:,getSamples(numSamples, sampleSize));
	w = (Xsub * ((Xsub' * w) .^ 2)) / size(Xsub, 2);
       case 43 then
	Xsub=X(:,getSamples(numSamples, sampleSize));
	EXGskew = (Xsub * ((Xsub' * w) .^ 2)) / size(Xsub, 2);
	Beta = w' * EXGskew;
	w = w - myy * (EXGskew - Beta*w)/(-Beta);
	
       else
	error('Code for desired nonlinearity not found!');
      end
      
      // Normalize the new w.
      w = w / norm(w);
      i = i + 1;
    end
    rnd = rnd + 1;
  end
  if b_verbose, printf('Done.\n'); end
  
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // Also plot the ones that may not have been plotted.
  if (usedDisplay > 0) & (rem(rnd-1, displayInterval) ~= 0)
    select usedDisplay
     case 1
      // There was and may still be other displaymodes...
      // 1D signals
      temp = X'*B;
      icaplot('dispsig',temp(:,1:numOfIC)');
      //drawnow;
     case 2
      // ... and now there are :-)
      // 1D basis
      icaplot('dispsig',A');
     // drawnow;
     case 3
      // ... and now there are :-)
      // 1D filters
      icaplot('dispsig',W);
      //drawnow;
     else
    end
  end
end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// In the end let's check the data for some security
if ~isreal(A)
  if b_verbose, printf('Warning: removing the imaginary part from the result.\n'); end
  A = real(A);
  W = real(W);
end



endfunction
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Subfunction
// Calculates tanh simplier and faster than Matlab tanh.
function out=rem(X,Y)
out=X-fix(X./Y).*Y
endfunction
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function y=my_tanh(x)
y = 1 - 2 ./ (exp(2 * x) + 1);
endfunction
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function Samples = getSamples(maxv, percentage)
Samples = find(rand(1, maxv) < percentage);
endfunction