function icaplot(pmode, varargin);
// plot signals in various ways
// Description
//
// icaplot is mainly for plottinf and comparing the mixed signals and
// separated ica-signals.
//
// icaplot has many different modes. The first parameter of the function
// defines the mode. Other parameters and their order depends on the
// mode. The explanation for the more common parameters is in the end.
//
// Classic
// 
//     icaplot('classic', s1, n1, range, xrange, titlestr)
//
//     Plots the signals in the same manner as the FASTICA and FASTICAG
//     programs do. All the signals are plotted in their own axis.
//
// Complot
// 
//     icaplot('complot', s1, n1, range, xrange, titlestr)
//
//     The signals are plotted on the same axis. This is good for
//     visualization of the shape of the signals. The scale of the signals 
//     has been altered so that they all fit nicely.
//
// Histogram
// 
//     icaplot('histogram', s1, n1, range, bins, style)
//     
//     The histogram of the signals is plotted. The number of bins can be
//     specified with 'bins'-parameter. The style for the histograms can
//     be either 'bar' (default) of 'line'.
//
// Scatter
// 
//     icaplot('scatter', s1, n1, s2, n2, range, titlestr, s1label,
//     s2label, markerstr)
//
//     A scatterplot is plotted so that the signal 1 is the 'X'-variable
//     and the signal 2 is the 'Y'-variable. The 'markerstr' can be used
//     to specify the maker used in the plot. The format for 'markerstr'
//     is the same as for Matlab's PLOT. 
//
// Compare
// 
//     icaplot('compare', s1, n1, s2, n2, range, xrange, titlestr,
//     s1label, s2label)
//
//     This for for comparing two signals. The main used in this context
//     would probably be to see how well the separated ICA-signals explain 
//     the observed mixed signals. The s2 signals are first scaled with
//     REGRESS function.
//
// Compare - Sum
// 
//     icaplot('sum', s1, n1, s2, n2, range, xrange, titlestr, s1label,
//     s2label)
//
//     The same as Compare, but this time the signals in s2 (specified by
//     n2) are summed together.
//
// Compare - Sumerror
// 
//     icaplot('sumerror', s1, n1, s2, n2, range, xrange, titlestr,
//     s1label, s2label)
//     
//     The same as Compare - Sum, but also the 'error' between the signal
//     1 and the summed IC's is plotted.
//
//
// More common parameters
// 
//     The signals to be plotted are in matrices s1 and s2. The n1 and n2
//     are used to tell the index of the signal or signals to be plotted
//     from s1 or s2. If n1 or n2 has a value of 0, then all the signals
//     from corresponding matrix will be plotted. The values for n1 and n2 
//     can also be vectors (like: [1 3 4]) In some casee if there are more
//     than 1 signal to be plotted from s1 or s2 then the plot will
//     contain as many subplots as are needed. 
//
//     The range of the signals to be plotted can be limited with
//     'range'-parameter. It's value is a vector ( 10000:15000 ). If range 
//     is 0, then the whole range will be plotted.
//
//     The 'xrange' is used to specify only the labels used on the
//     x-axis. The value of 'xrange' is a vector containing the x-values
//     for the plots or [start end] for begin and end of the range
//     ( 10000:15000 or [10 15] ). If xrange is 0, then value of range
//     will be used for x-labels.
//
//     You can give a title for the plot with 'titlestr'. Also the
//     's1label' and 's2label' are used to give more meaningfull label for 
//     the signals.
//
//     Lastly, you can omit some of the arguments from the and. You will
//     have to give values for the signal matrices (s1, s2) and the
//     indexes (n1, n2)
//     Authors
//     H. Nahrstaedt - 2010
//     (C) 1996-2005 by Hugo Gävert, Jarmo Hurri, Jaakko Särelä, and Aapo Hyvärinen. 

// @(#)$Id: icaplot.m,v 1.2 2003/04/05 14:23:58 jarmo Exp $

if pmode=='classic' then
     pmode='';
elseif pmode=='dispsig'
     pmode='';
end;

smode=pmode;
if pmode=='sum' then
    pmode='compare';
elseif pmode=='sumerror' then
    pmode='compare';
end;

select pmode
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // 'dispsig' is to replace the old DISPSIG
  // '' & 'classic' are just another names - '' quite short one :-)
 //case {'', 'classic', 'dispsig'} 
 case '' then
  // icaplot(mode, s1, n1, range, xrange, titlestr)
  if length(varargin) < 1, error('Not enough arguments.'); end
  if length(varargin) < 5, titlestr = '';else titlestr = varargin(5); end
  if length(varargin) < 4, xrange = 0;else xrange = varargin(4); end
  if length(varargin) < 3, rrange = 0;else rrange = varargin(3); end
  if length(varargin) < 2, n1 = 0;else n1 = varargin(2); end
  s1 = varargin(1);
  rrange=chkrange(rrange, s1);
  xrange=chkxrange(xrange, rrange);
  n1=chkn(n1, s1);

  clf;
  
  numSignals = size(n1, 2);
  for i = 1:numSignals,
    subplot(numSignals, 1, i);
    plot(xrange, s1(n1(i), rrange));
  end
  subplot(numSignals,1, 1);
  if (~isempty(titlestr))
    title(titlestr);
  end
  
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 case 'complot' then
  // icaplot(mode, s1, n1, range, xrange, titlestr)
  if length(varargin) < 1, error('Not enough arguments.'); end
  if length(varargin) < 5, titlestr = '';else titlestr = varargin(5); end
  if length(varargin) < 4, xrange = 0;else xrange = varargin(4); end
  if length(varargin) < 3, rrange = 0;else rrange = varargin(3); end
  if length(varargin) < 2, n1 = 0;else n1 = varargin(2); end
  s1 = remmean(varargin(1));
  rrange=chkrange(rrange, s1);
  xrange=chkxrange(xrange, rrange);
  n1=chkn(n1, s1);
  
  for i = 1:size(n1, 2)
    S1(i, :) = s1(n1(i), rrange);
  end
  
  alpha = mean(max(S1')-min(S1'));
  for i = 1:size(n1,2)
    S2(i,:) = S1(i,:) - alpha*(i-1)*ones(S1(1,:));
  end
  
  plot(xrange, S2');
  //axis([min(xrange) max(xrange) min(min(S2)) max(max(S2)) ]);
  a=gca();
  a.data_bounds=[min(xrange) max(xrange) min(min(S2)) max(max(S2)) ];
  //set(gca,'YTick',(-size(S1,1)+1)*alpha:alpha:0);
  //set(gca,'YTicklabel',fliplr(n1));
  
  if (~isempty(titlestr))
    title(titlestr);
  end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 case 'histogram' then
  // icaplot(mode, s1, n1, range, bins, style)
  if length(varargin) < 1, error('Not enough arguments.'); end
  if length(varargin) < 5, style = 'bar';else style = varargin(5); end
  if length(varargin) < 4, bins = 10;else bins = varargin(4); end
  if length(varargin) < 3, rrange = 0;else rrange = varargin(3); end
  if length(varargin) < 2, n1 = 0;else n1 = varargin(2); end
  s1 = varargin(1);
  rrange = chkrange(rrange, s1);
  n1 = chkn(n1, s1);
  
  numSignals = size(n1, 2);
  rows = floor(sqrt(numSignals));
  columns = ceil(sqrt(numSignals));
  while (rows * columns < numSignals)
    columns = columns + 1;
  end
  
  select style
   case '' then
    for i = 1:numSignals,
      subplot(rows, columns, i);
      nan_hist(s1(n1(i), rrange), bins,'plot');
      title(string(n1(i)));
      //drawnow;
    end
    case 'bar' then
    for i = 1:numSignals,
      subplot(rows, columns, i);
      nan_hist(s1(n1(i), rrange), bins,'plot');
      title(string(n1(i)));
      //drawnow;
    end
   case 'line' then
    for i = 1:numSignals,
      subplot(rows, columns, i);
      [Y, X]=nan_hist(s1(n1(i), rrange), bins);
      plot(X, Y);
      title(string(n1(i)));
      //drawnow;
    end
   else
    printf('Unknown style.\n')
  end

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 case 'scatter' then
  // icaplot(mode, s1, n1, s2, n2, range, titlestr, xlabelstr, ylabelstr, markerstr)
  if length(varargin) < 4, error('Not enough arguments.'); end
  if length(varargin) < 9, markerstr = '.';else markerstr = varargin(9); end
  if length(varargin) < 8, ylabelstr = 'Signal 2';else ylabelstr = varargin(8); end
  if length(varargin) < 7, xlabelstr = 'Signal 1';else xlabelstr = varargin(7); end
  if length(varargin) < 6, titlestr = '';else titlestr = varargin(6); end
  if length(varargin) < 5, rrange = 0;else rrange = varargin(5); end
  n2 = varargin(4);
  s2 = varargin(3);
  n1 = varargin(2);
  s1 = varargin(1);
  rrange = chkrange(rrange, s1);
  n1 = chkn(n1, s1);
  n2 = chkn(n2, s2);
  
  rows = size(n1, 2);
  columns = size(n2, 2);
  for r = 1:rows
    for c = 1:columns
      subplot(rows, columns, (r-1)*columns + c);
      plot(s1(n1(r), rrange),s2(n2(c), rrange),markerstr);
      e=gce();
      e.children(1).mark_size=1;

      if (~isempty(titlestr))
	title(titlestr);
      end
      if (rows*columns == 1)
	xlabel(xlabelstr);
	ylabel(ylabelstr);
      else 
	xlabel(xlabelstr +' ('+ string(n1(r)) +')');
	ylabel(ylabelstr +' ('+ string(n2(c)) +')');
      end
     // drawnow;
    end
  end
  
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 //case {'compare', 'sum', 'sumerror'} then
 case 'compare' then
  // icaplot(mode, s1, n1, s2, n2, range, xrange, titlestr, s1label, s2label)
  if length(varargin) < 4, error('Not enough arguments.'); end
  if length(varargin) < 9, s2label = 'IC';else s2label = varargin(9); end
  if length(varargin) < 8, s1label = 'Mix';else s1label = varargin(8); end
  if length(varargin) < 7, titlestr = '';else titlestr = varargin(7); end
  if length(varargin) < 6, xrange = 0;else xrange = varargin(6); end
  if length(varargin) < 5, rrange = 0;else rrange = varargin(5); end
  s1 = varargin(1);
  n1 = varargin(2);
  s2 = varargin(3);
  n2 = varargin(4);
  rrange = chkrange(rrange, s1);
  xrange = chkxrange(xrange, rrange);
  n1 = chkn(n1, s1);
  n2 = chkn(n2, s2);

  numSignals = size(n1, 2);
  if (numSignals > 1)
    externalLegend = 1;
  else
    externalLegend = 0;
  end
  
  rows = floor(sqrt(numSignals+externalLegend));
  columns = ceil(sqrt(numSignals+externalLegend));
  while (rows * columns < (numSignals+externalLegend))
    columns = columns + 1;
  end
  
  clf;
  
  for j = 1:numSignals
    subplot(rows, columns, j);
    select smode
     case 'compare' then
      plotcompare(s1, n1(j), s2,n2, rrange, xrange);
      [legendtext,legendstyle]=legendcompare(n1(j),n2,s1label,s2label,externalLegend);
     case 'sum' then
      plotsum(s1, n1(j), s2,n2, rrange, xrange);
      [legendtext,legendstyle]=legendsum(n1(j),n2,s1label,s2label,externalLegend);
     case 'sumerror' then
      plotsumerror(s1, n1(j), s2,n2, rrange, xrange);
      [legendtext,legendstyle]=legendsumerror(n1(j),n2,s1label,s2label,externalLegend);
    end
    
    if externalLegend
      title(titlestr+ ' ('+ s1label + ' ' +string(n1(j)) +')');
    else
      legend(char(legendtext));
      if (~isempty(titlestr))
	title(titlestr);
      end
    end
  end
  
  if (externalLegend)
    subplot(rows, columns, numSignals+1);
    legendsize = length(legendtext);
    //hold on;
    for i=1:legendsize
      plot([0 1],[legendsize-i legendsize-i], char(legendstyle(i)));
      //text(1.5, legendsize-i, char(legendtext(i)));
      xstring(1.5, legendsize-i,char(legendtext(i)));
    end
    //hold off;
    //axis([0 6 -1 legendsize]);
    a=gca();
    a.data_bounds=[0 6 -1 legendsize];
    //axis off;
  end
  

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
end;
endfunction
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function plotcompare(s1, n1, s2, n2, rrange, xrange);
  style=getStyles();
  //K = regress(s1(n1,:)',s2');
  K = s1(n1,:)'\s2';
  plot(xrange, s1(n1,rrange), char(style(1)));
  //hold on
  for i=1:size(n2,2)
    plotstyle=char(style(i+1));
    plot(xrange, K(n2(i))*s2(n2(i),rrange), plotstyle);
  end
 //hold off
endfunction
function [legendText, legendStyle]=legendcompare(n1, n2, s1l, s2l, externalLegend);
  style=getStyles();
  legendText=list();
  if (externalLegend)
    legendText(1)={s1l +' (see the titles)'};
  else
    legendText(1)={s1l +' '+ string(n1)};
  end
  legendStyle(1)=style(1);
  for i=1:size(n2, 2)
    legendText(i+1) = {s2l+ ' '+ string(n2(i))};
    legendStyle(i+1) = style(i+1);
  end
endfunction
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function plotsum(s1, n1, s2, n2, rrange, xrange);
  //K = diag(regress(s1(n1,:)',s2'));
  K = diag(s1(n1,:)'\s2');
  sigsum = mtlb_sum(K(:,n2)*s2(n2,:));
  plot(xrange, s1(n1, rrange),'k-', ...
       xrange, sigsum(rrange), 'b-');
endfunction
function [legendText, legendStyle]=legendsum(n1, n2, s1l, s2l, externalLegend);
  legendText=list();
  if (externalLegend)
    legendText(1)={s1l +' (see the titles)'};
  else
    legendText(1)={s1l +' '+ string(n1)};
  end
  legendText(2)={'Sum of '+ s2l+ ': '+string(n2)};
  legendStyle={'k-';'b-'};
endfunction
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function plotsumerror(s1, n1, s2, n2, rrange, xrange);
  //K = diag(regress(s1(n1,:)',s2'));
  K = diag(s1(n1,:)'\s2')
  sigsum = mtlb_sum(K(:,n2)*s2(n2,:));
  plot(xrange, s1(n1, rrange),'k-');
  plot(xrange, sigsum(rrange), 'b-');
  plot(xrange, s1(n1, rrange)-sigsum(rrange), 'r-');
endfunction
function [legendText, legendStyle]=legendsumerror(n1, n2, s1l, s2l, externalLegend);
  legendText=list();
  if (externalLegend)
    legendText(1)={s1l +' (see the titles)'};
  else
    legendText(1)={s1l+ ' '+ string(n1)};
  end
  legendText(2)={'Sum of '+ s2l+ ': '+ string(n2)};
  legendText(3)={'''Error'''};
  legendStyle={'k-';'b-';'r-'};
endfunction
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function style=getStyles;
  col = {'k','r','g','b','m','c','y'};
  line = {'-',':','-.','--'};
  for i = 0:size(line,2)-1
    for j = 1:size(col, 2)
      style(j + i*size(col, 2)) = strcat(col(j), line(i+1));
    end
  end
endfunction
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function rrange=chkrange(r, s)
  if r == 0
    rrange = 1:size(s, 2);
  else
    rrange = r;
  end
endfunction
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function xrange=chkxrange(xr,r);
  if xr == 0
    xrange = r;
  elseif size(xr, 2) == 2
    xrange = xr(1):(xr(2)-xr(1))/(size(r,2)-1):xr(2);
  elseif size(xr, 2)~=size(r, 2)
    error('Xrange and range have different sizes.');
  else
    xrange = xr;
  end
endfunction
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function n=chkn(n,s)
  if n == 0,
    n = 1:size(s, 1);
  end
endfunction