function [newVectors, whiteningMatrix, dewhiteningMatrix] = whitenv  (vectors, E, D, s_verbose);
//Whitenv vectors.
// Calling Sequence
// [newVectors, whiteningMatrix, dewhiteningMatrix] =   whitenv(vectors, E, D, verbose);
//
// Parameters
//
// vectors  :     Data in row vectors.
// E        :     Eigenvector matrix from function 'pcamat'
// D        :     Diagonal eigenvalue matrix from function 'pcamat'
// verbose  :     Optional. Default is 'on'
//
// Examples
//       [E, D] = pcamat(vectors);
//       [nv, wm, dwm] = whitenv(vectors, E, D);
//
// Description
// Whitens the data (row vectors) and reduces dimension. Returns
// the whitened vectors (row vectors), whitening and dewhitening matrices.
// 
//
//   See also 
//   pcamat
//   Authors
//   H. Nahrstaedt - 2011
//   (C) 1996-2005 by Hugo Gävert, Jarmo Hurri, Jaakko Särelä, and Aapo Hyvärinen. 

// @(#)$Id: whitenv.m,v 1.3 2003/10/12 09:04:43 jarmo Exp $

 [nargout,nargin]=argn();
// ========================================================
// Default value for 'verbose'
if nargin < 4, s_verbose = 'on'; end

// Check the optional parameter verbose;
select convstr(s_verbose,"l")
 case 'on' then
  b_verbose = 1;
 case 'off' then
  b_verbose = 0;
 else
  error(sprintf('Illegal value [ %s ] for parameter: ''verbose''\n', s_verbose));
end

// ========================================================
// In some cases, rounding errors in Matlab cause negative
// eigenvalues (elements in the diagonal of D). Since it
// is difficult to know when this happens, it is difficult
// to correct it automatically. Therefore an error is 
// signalled and the correction is left to the user.
if or (diag (D) < 0),
  error (sprintf (['[ %d ] negative eigenvalues computed from the' ...
		   ' covariance matrix.\nThese are due to rounding' ...
		   ' errors in Matlab (the correct eigenvalues are\n' ...
		   'probably very small).\nTo correct the situation,' ...
		   ' please reduce the number of dimensions in the' ...
		   ' data\nby using the ''lastEig'' argument in' ...
		   ' function FASTICA, or ''Reduce dim.'' button\nin' ...
		   ' the graphical user interface.'], ...
		  sum (diag (D) < 0)));
end

// ========================================================
// Calculate the whitening and dewhitening matrices (these handle
// dimensionality simultaneously).
whiteningMatrix = inv (sqrt (D)) * E';
dewhiteningMatrix = E * sqrt (D);

// Project to the eigenvectors of the covariance matrix.
// Whiten the samples and reduce dimension simultaneously.
if b_verbose, printf ('Whitening...\n'); end
newVectors =  whiteningMatrix * vectors;

// ========================================================
// Just some security...
if ~isreal(newVectors)
  error ('Whitened vectors have imaginary values.');
end

// Print some information to user
if b_verbose
  printf ('Check: covariance differs from identity by [ %g ].\n',  max (abs (mtlb_cov (newVectors', 1) - eye (size (newVectors, 1),size (newVectors, 1)))));
end
endfunction